"""GELF log format support."""
import json
import time
from typing import Dict, Protocol

from eventail.log_criticity import CRITICITY_LABELS


class Endpoint(Protocol):
    HOSTNAME: str
    ID: int
    logical_service: str


class GELF:
    def __init__(
        self,
        endpoint: Endpoint,
        criticity: int,
        short: str,
        full: str = "",
        conversation_id: str = "",
        additional_fields: Dict = {},
    ) -> None:
        level_name = CRITICITY_LABELS[criticity % 8]
        log = {
            "version": "1.1",
            "short_message": short,
            "full_message": full,
            "level": criticity,
            "_levelname": level_name,
            "host": f"{endpoint.logical_service}@{endpoint.HOSTNAME}",
            "timestamp": time.time(),
            "_conversation_id": conversation_id,
            "_logical_service": endpoint.logical_service,
            "_worker_pid": endpoint.ID,
        }
        for key, value in additional_fields.items():
            assert key != "id" and key != "_id", "GELF: 'id' can't be used as key name!"
            if key.startswith("_"):  # already escaped
                log[key] = value
            else:
                log[f"_{key}"] = value
        self._log = log
        self._routing_key = "{}.{}".format(endpoint.logical_service, level_name)

    @property
    def payload(self) -> bytes:
        return json.dumps(self._log).encode("utf-8")

    @property
    def routing_key(self) -> str:
        return self._routing_key



#
# MIT License
#
# Copyright (c) 2021 Groupe Allo-Media
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Process incoming events in batches."""


from dataclasses import dataclass
from typing import Any, Callable, List, Optional, Union

from eventail.async_service.aio import JSON_MODEL
from eventail.async_service.aio import Service as AioService
from eventail.async_service.pika import Service as PikaService

Service = Union[AioService, PikaService]


@dataclass
class RawEvent:
    payload: JSON_MODEL
    conversation_id: str
    meta: JSON_MODEL


class Batch:
    """A class for batch processing of events.

    You can instantiate as many Batch as you need, but only one thread can update it.
    You can't use Batch for Commands or Results.
    """

    def __init__(
        self,
        endpoint: Service,
        max_entries: int,
        timeout: float,
        callback: Callable[[List[RawEvent]], None],
    ) -> None:
        """Create a new Batch Processor.

          - `max_entries` is the maximum number of events to accumulate before triggering the batch;
          - `timeout` is the maximum time in seconds to wait for a complete batch after the first event
            is pushed;
          - `callback` is a callable that will be called with a list of `RawEvent`
            as arguments when the batch is triggered.

        Once you have a batch processor instance, all you have to do is to `.push()` events to it and it
        takes care of the rest.

        You can manually trigger the batch by calling the `.expire()` method.
        """
        self.endpoint = endpoint
        self.max_entries = max_entries
        self.timeout = timeout
        self.callback = callback
        self._timer: Optional[Any] = None
        self._batch: List[RawEvent] = []

    def push(self, payload: JSON_MODEL, conversation_id: str, meta: JSON_MODEL):
        """Push a new event to the batch."""
        if not self._batch:
            self._timer = self.endpoint.call_later(self.timeout, self._expired)
        self._batch.append(RawEvent(payload, conversation_id, meta))
        if len(self._batch) == self.max_entries:
            self.expire()

    def expire(self):
        if self._timer is not None:
            self.endpoint.cancel_timer(self._timer)
        self._expired()

    def _expired(self):
        self._timer = None
        try:
            self.callback(self._batch)
        finally:
            self._batch.clear()



from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='eventail',
    version='2.0.0',
    url='https://github.com/allo-media/eventail',
    author='Allo-Media',
    author_email='dev@allo-media.fr',
    description='A base class and utilities for AM service architecture',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
        "Intended Audience :: Developers",
        "Operating System :: POSIX :: Linux",
        "Topic :: Software Development",
    ],
    license='MIT',
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    scripts=["scripts/logger.py", "scripts/monitor.py", "scripts/publish_event.py", "scripts/send_command.py",
             "scripts/inspect_queue.py", "scripts/resurrect.py"],
    install_requires=[
        "pika>=1.2.0",
        "cbor",
        "redis"
    ],
    extras_require={
        'asyncio': ["aiormq", "uvloop"],
        'synchronous': ["kombu"],
        'test': ["tox"]
    },
    python_requires='>=3.8',
)



#
# MIT License
#
# Copyright (c) 2018-2019 Groupe Allo-Media
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
A base class implementing AM service architecture and its requirements for a synchronous publisher Endpoint.
"""
import logging
import os
import socket
from typing import Any, Dict, List, Optional

import cbor
from kombu import Connection, Exchange
from kombu.pools import producers, set_limit

from eventail.gelf import GELF

JSON_MODEL = Dict[str, Any]
LOGGER = logging.getLogger("sync_endpoint")
set_limit(2)  # two connections are more than enough; `async_service` uses just one.


class Endpoint:
    """A synchronous publishing endpoint for AlloMedia EDA."""

    ID = os.getpid()
    HOSTNAME = socket.gethostname()
    EVENT_EXCHANGE = "events"
    LOG_EXCHANGE = "logs"
    EVENT_EXCHANGE_TYPE = "topic"
    LOG_EXCHANGE_TYPE = "topic"

    def __init__(
        self,
        amqp_urls: List[str],
        logical_service: str,
        connection_max_retries: Optional[int] = None,
    ) -> None:
        """Initialize endpoint.

        * ``amqp_urls`` is a list of broker urls that will be tried in turn (round robin style) to send messages.
        * ``logical_service``: the logical service this endpoint belongs to.
        """

        self.logical_service = logical_service

        self._connection = Connection(
            amqp_urls,
            transport_options={
                "confirm_publish": True,
                "max_retries": connection_max_retries,
            },
        )
        self.event_exchange = Exchange(
            self.EVENT_EXCHANGE, type=self.EVENT_EXCHANGE_TYPE, durable=True
        )
        self.log_exchange = Exchange(
            self.LOG_EXCHANGE, self.LOG_EXCHANGE_TYPE, durable=True
        )
        self._force_json = False

    def force_json(self):
        """Force serialization of payload into JSON."""
        self._force_json = True

    def log(
        self,
        criticity: int,
        short: str,
        full: str = "",
        conversation_id: str = "",
        additional_fields: Dict = {},
    ) -> None:
        """Log to the log bus.


        Parameters:
         - `criticity`: int, in the syslog scale
         - `short`: str, short description of log
         - `full`: str, the full message of the log (appears as `message` in Graylog)
         - `additional_fields: Dict, data to be merged into the GELF payload as additional fields
        """
        # no persistent messages, no retry
        gelf = GELF(self, criticity, short, full, conversation_id, additional_fields)
        LOGGER.debug("Application logged: %s\n%s", short, full)

        # the timeout here has no measurable effect in my tests, but I set it here to rule out
        # possible issues
        with producers[self._connection].acquire(block=True, timeout=2) as producer:
            producer.publish(
                gelf.payload,
                delivery_mode=1,  # not persistent
                exchange=self.log_exchange,
                serializer=None,
                routing_key=gelf.routing_key,
                declare=[self.log_exchange],
                retry=False,
            )

    def publish_event(
        self,
        event: str,
        message: JSON_MODEL,
        conversation_id: str,
        max_retries: Optional[int] = None,
    ) -> None:
        """Publish an event on the bus.

        The ``event`` is the name of the event,
        and the `message` is any data conforming to the JSON model.
        `max_retries` is None by default (retry for ever) but can be an `int`
        to specify the number of attempts to send the message before the error is raised.
        The firt retry is immediate, then the interval increases by one second at each step.
        """

        # the timeout here has no measurable effect in my tests, but I set it here to rule out
        # possible issues
        with producers[self._connection].acquire(block=True, timeout=2) as producer:
            headers = {"conversation_id": str(conversation_id)}
            producer.publish(
                message if self._force_json else cbor.dumps(message),
                delivery_mode=2,  # persistent
                exchange=self.event_exchange,
                routing_key=event,
                declare=[self.event_exchange],
                headers=headers,
                content_type=None if self._force_json else "application/cbor",
                content_encoding=None if self._force_json else "binary",
                serializer="json" if self._force_json else None,
                retry=True,
                retry_policy={
                    "interval_start": 0,  # First retry immediately,
                    "interval_step": 1,  # then increase by 1s for every retry.
                    "interval_max": 30,  # but don't exceed 30s between retries.
                    "max_retries": max_retries,
                },
            )
